# Copyright (c) 2024 DomainTools LLC
# This work is unpublished, and unauthorized reproduction is prohibited.
# DomainTools, LLC owns all rights to this work and intends that
# it be held confidential, preserved as a trade secret, and protected as
# an unpublished copyrighted work.

import os
import logging
from time import sleep

from .util import get_open_files
from .fileset import parse_part_file, File, compute_overlap, ParseError

logger = logging.getLogger(__name__)


class Cleaner:
    def __init__(self, fileset, download_manager, interval=None, minimal=True):
        self.fileset = fileset
        self.download_manager = download_manager
        self.interval = interval
        self.minimal = minimal

    def start(self):
        if self.interval is None:
            logging.error("Cleaner for fileset %s started with None interval", self.fileset.uri)
            return
        elif not isinstance(self.interval, int):
            logging.error("Cleaner for fileset %s started with bad interval %s", self.fileset.uri, self.interval)
            return
        elif self.interval < 1:
            logging.error("Cleaner for fileset %s started with less than 1 second interval %s", self.fileset.uri,
                          self.interval)
            return

        while True:
            sleep(self.interval)
            logger.debug("Cleaner for fileset %s awoken from sleep", self.fileset.uri)
            self._clean()

    def _clean(self):
        open_files = get_open_files()
        local_fileset = self.fileset.get_local_fileset()
        try:
            remote_fileset = self.fileset.get_remote_fileset()
        except Exception as e:
            logger.error("Cleaner encountered this error fetching the remote fileset: %s", e)
            return

        for tf in self.fileset.list_temporary_files():
            if tf in open_files:
                continue

            try:
                f = File(parse_part_file(os.path.basename(tf)), dname=self.fileset)
            except ParseError:
                logger.error("Cleaner failed to parse %s. Ignoring.", tf)
                continue

            if f in self.download_manager:
                continue

            # First check if it not behind the remote fileset, in other words, it is within or ahead of the remote
            if not f.is_behind(remote_fileset):
                if self.minimal:
                    lf_copy = local_fileset.copy()
                    lf_copy.add(f)
                    # Then if it is not overlapped by whats on disk, keep it
                    if f not in list(compute_overlap(lf_copy)):
                        continue
                else:
                    # A full fileset means that overlaps are allowed, so keep this part file
                    continue

            logger.debug("Cleaner unlinking tempfile: %s", tf)
            try:
                os.unlink(tf)
            except Exception:
                # It might be in the process of being deleted or was deleted already by another thread
                logger.error("Cleaner encountered an error unlinking: %s", tf)
