# Copyright (c) 2022-2024 DomainTools LLC
# Copyright (c) 2015-2016, 2019, 2021-2022 by Farsight Security, Inc.
# This work is unpublished, and unauthorized reproduction is prohibited.
# DomainTools, LLC owns all rights to this work and intends that
# it be held confidential, preserved as a trade secret, and protected as
# an unpublished copyrighted work.

import heapq
import logging
import os
import time
import threading
import traceback
import urllib.request
import urllib.error
import urllib.parse
from email.utils import parsedate_tz, mktime_tz
import base64

from .digest import digest_extension, DigestChecker, DigestError
from .util import iterfileobj, TempFile, get_total_content_length, TerminatableThread

logger = logging.getLogger(__name__)


class DownloadError(Exception):
    pass


class DownloadManager:
    def __init__(self, max_downloads=4, download_timeout=None, retry_timeout=60):
        self._pending_downloads = set()
        self._active_downloads = dict()

        # A dictionary where the keys are File(s) that failed to download. Being in this dict prevents a File from
        # being added to _pending_downloads until the retry_timeout has elapsed.
        self._failed_downloads = dict()

        self._max_downloads = max_downloads
        self._download_timeout = download_timeout
        self._retry_timeout = retry_timeout
        self._lock = threading.RLock()

        self._main_thread = None
        self._action_required = threading.Condition()
        self._terminate = threading.Event()

    def start(self):
        logger.debug('Starting DownloadManager {}'.format(self))
        if self._main_thread:
            raise Exception('already running')

        self._terminate.clear()

        self._main_thread = threading.Thread(target=self._run, daemon=True)
        self._main_thread.start()

    def stop(self, blocking=False, timeout=None):
        logger.debug('Stopping DownloadManager {}'.format(self))
        self._terminate.set()
        with self._action_required:
            logger.debug('Notifying all run loops')
            self._action_required.notify_all()
        if blocking or timeout:
            return self.join(timeout=timeout)

    def join(self, timeout=None):
        logger.debug('Joining DownloadManager {}'.format(self))
        return self._main_thread.join(timeout=timeout)

    def _run(self):
        logger.debug('Running DownloadManager {}'.format(self))
        while not self._terminate.is_set():
            with self._lock:
                for f,thread in list(self._active_downloads.items()):
                    if not thread.is_alive():
                        del self._active_downloads[f]
                        logger.debug('Joining {}'.format(thread))
                        thread.join()
                for f,thread in list(self._failed_downloads.items()):
                    if not thread.is_alive():
                        del self._failed_downloads[f]
                        logger.debug('Joining {}'.format(thread))
                        thread.join()

            with self._lock:
                for f in heapq.nlargest(self._max_downloads - len(self._active_downloads), self._pending_downloads):
                    self._pending_downloads.remove(f)

                    thread = TerminatableThread(target=self._download, args=(f,), daemon=True)
                    thread.start()

                    self._active_downloads[f] = thread

            with self._action_required:
                logger.debug('Waiting DownloadManager {}'.format(self))
                self._action_required.wait()
                logger.debug('Awoken DownloadManager {}'.format(self))

        logger.debug('Completing DownloadManager run {}'.format(self))
        with self._lock:
            for f,thread in list(self._active_downloads.items()):
                if thread.is_alive():
                    thread.terminate()
                thread.join()
                del self._active_downloads[f]
            for f,thread in list(self._failed_downloads.items()):
                if thread.is_alive():
                    thread.terminate()
                thread.join()
                del self._failed_downloads[f]

    def _download(self, f):
        logger.debug('Downloading %s', f)
        try:
            target = f.target()

            out = f.get_download_file()
            digest_checker = None

            resume_pos = out.tell()
            if resume_pos == 0:
                logger.info('Downloading %s to %s', f.uri, target)
            else:
                logger.info('Resuming download of %s to %s from %s', f.uri, target, resume_pos)

            req = urllib.request.Request(f.uri)
            if f.apikey:
                req.add_header('X-API-Key', f.apikey)
            if resume_pos > 0:
                logger.debug('Requesting Range: bytes=%s-', resume_pos)
                req.add_header('Range', 'bytes={}-'.format(resume_pos))
            fp = urllib.request.urlopen(req, timeout=self._download_timeout)

            algorithm, digest, digest_file = None, None, None
            if 'Digest' in fp.headers:
                algorithm, _, digest = fp.headers['Digest'].partition('=')
                digest_file = '{}.{}'.format(target, digest_extension(algorithm))
            elif f.digest_required:
                raise DownloadError('Digest header missing and digest_required=True')

            new_data = threading.Event()
            if algorithm is not None:
                digest_checker = DigestChecker(algorithm)
                digest_checker.start_fill(out.name, new_data)

            logger.debug('Copying urlopen of %s to %s', f.uri, out.name)
            for chunk in iterfileobj(fp):
                out.write(chunk)
                new_data.set()

            cur_pos = out.tell()
            correct_length = False
            if 'Content-Length' in fp.headers:
                try:
                    received_len = int(fp.headers['Content-Length'])

                    expected_len = get_total_content_length(received_len, resume_pos, fp.headers['Content-Range'])

                    if cur_pos != expected_len:
                        raise DownloadError('Content length mismatch: file length {} != content length {}'.format(
                            cur_pos, expected_len))
                    correct_length = True
                except ValueError:
                    logger.debug('Skipping content length check, invalid header: %s',
                                 fp.headers['Content-Length'])
            else:
                logger.debug('Skipping content length check, header missing')

            if digest_checker is not None:
                logger.info("Validating download from %s", f.uri)
                digest_pos = digest_checker.stop_fill()
                if digest_pos < cur_pos:
                    out.seek(digest_pos)
                    start = time.time()
                    logger.debug("Continuing Digest Check...")
                    for chunk in iterfileobj(out):
                        digest_checker.update(chunk)
                    logger.debug("Finished Digest Check. Took %s seconds.", round(time.time() - start, 2))
                    # Check that digest_checker has all the expected input
                    assert (out.tell() == cur_pos)

                digest_checker.check(digest)
            else:
                logger.debug('Skipping digest check')

            logger.debug('Completed download')

            out.close()
            os.chmod(out.name, 0o644)

            mtime_tz = fp.headers.get('Last-Modified', None)
            if mtime_tz:
                mtime = mktime_tz(parsedate_tz(mtime_tz))
                logger.debug('Setting mtime of %s to %s', out.name, time.ctime(mtime))
                os.utime(out.name, (mtime, mtime))

            f.validate(out.name)

            if digest_file:
                logger.debug('Writing digest=%s to %s', digest, digest_file)
                with TempFile(prefix='.{}.'.format(os.path.basename(digest_file)), dir=f.dname, destination=digest_file) as tf:
                    tf.write('{}  {}\n'.format(base64.b64decode(digest).hex(), os.path.basename(target)).encode('utf-8'))

            try:
                logger.debug('Renaming %s to %s', out.name, target)
                os.rename(out.name, target)
                out.delete = False
            except Exception:
                try:
                    os.unlink(digest_file)
                except OSError:
                    pass
                raise

            logger.info('Download of %s to %s complete', f.uri, target)
        except urllib.error.HTTPError as e:
            if e.code == 416:
                logger.debug('%s is larger than expected. Deleting and retrying.', out.name)
                os.unlink(out.name)
            self._handle_download_exception(e, f, out, digest_checker)
        except DigestError as e:
            if correct_length:
                logger.debug('%s has correct length but bad checksum. Deleting and retrying.',
                             out.name)
                os.unlink(out.name)
            self._handle_download_exception(e, f, out, digest_checker)
        except Exception as e:
            self._handle_download_exception(e, f, out, digest_checker)
        finally:
            with self._action_required:
                logger.debug('Notifying run loop')
                self._action_required.notify()

    def _handle_download_exception(self, e, f, out, dc):
        logger.error('Download of %s failed: %s', f.uri, str(e))
        logger.debug(traceback.format_exc())

        expire_thread = TerminatableThread(target=self._expire_failed_download, args=(f,), daemon=True)
        expire_thread.start()
        with self._lock:
            self._failed_downloads[f] = expire_thread
        if out and not out.closed:
            out.close()
        if dc:
            dc.stop_fill()

    def _expire_failed_download(self, f, timeout=None):
        # A sleep thread that just keeps the given File(f) in a dict so that it is not retried until the timeout has elapsed
        if timeout is None:
            timeout = self._retry_timeout
        logger.debug('Waiting {timeout} to retry {uri}'.format(timeout=timeout, uri=f.uri))
        time.sleep(timeout)
        logger.info('Failure timeout for {uri} complete'.format(uri=f.uri))
        with self._action_required:
            logger.debug('Notifying run loop')
            self._action_required.notify()

    def __contains__(self, filename):
        with self._lock:
            return filename in self._pending_downloads or filename in self._active_downloads or filename in self._failed_downloads

    def reconcile_missing(self, missing):
        new_pending = set()
        with self._lock:
            for f in missing:
                if f not in self._active_downloads and f not in self._failed_downloads:
                    new_pending.add(f)

            self._pending_downloads = new_pending

        with self._action_required:
            logger.debug('Notifying run loop')
            self._action_required.notify()


