# Copyright (c) 2022-2024 DomainTools LLC
# Copyright (c) 2015-2016, 2019, 2021 by Farsight Security, Inc.
# This work is unpublished, and unauthorized reproduction is prohibited.
# DomainTools, LLC owns all rights to this work and intends that
# it be held confidential, preserved as a trade secret, and protected as
# an unpublished copyrighted work.

import tempfile
import os
import threading
import ctypes
import psutil

IO_LENGTH = 16 * 1024

def get_total_content_length(content_length, tell, content_range=None):
    """
    Try to get the total content length from the content_range header string otherwise
    just use the calculated value
    """
    if content_range is not None:
        try:
            return int(content_range.split('/')[1])
        except (ValueError, IndexError):
            pass
    # Calculate the value
    return content_length + tell

def iterfileobj(fp, length=IO_LENGTH):
    '''iterate data from file-like object fp, yields bytes'''
    while 1:
        buf = fp.read(length)
        if not buf:
            break
        yield buf


def merge_dicts(dicts):
    """
    Merges a list dicts with values from the subsequent dict overwriting values from the previous dict that have matching keys
    :param dicts: The list of dictionaries
    :return: A dictionary
    """
    def update(d, u):
        for k, v in u.items():
            if isinstance(v, dict) and (d.get(k) is not None and isinstance(d.get(k), dict)):
                d[k] = update(d.get(k), v)
            else:
                d[k] = v
        return d

    z = None
    for d in dicts:
        if z is None:
            z = d.copy()   # start with the first dictionary's keys and values
        else:
            z = update(z, d)
    return z

class TempFile:
    """
    Returns a file-like object used to write fileset, mtbl, or digest files. The recommended use is with a context manager.
    If the destination parameter is given, the created file will be renamed to that when the runtime context exits.
    :param prefix: a string prefix for the temporary file
    :param dir: the directory where you want the temporary file to be created
    :param mode: an optional mode for the destination file (default: 0o644)
    :param destination: an optional path for the temp file to be renamed to when the runtime context exits.

    Example Usage:
        with TempFile(prefix='my-temp-file-', dir='/home/foo/Downloads', destination='/home/foo/export/rdata.fileset') as tf:
            tf.write(b'rdata.20210323.1736.m.tgdb\n')
    """
    def __init__(self, prefix, dir, mode=0o644, destination=None, delete=True):
        self._tempfile = tempfile.NamedTemporaryFile(prefix=prefix, dir=dir, delete=False)
        self.mode = mode
        self.destination = destination
        self.name = self._tempfile.name
        self.delete = delete

    def __getattr__(self, name):
        return self._tempfile.__getattr__(name)

    def __del__(self):
        if self.delete:
            try:
                os.unlink(self.name)
            except OSError:
                pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type:
            self.__del__()
            return False  # Continue propagating the exception

        try:
            self._tempfile.close()
            os.chmod(self.name, self.mode)
            if self.destination is not None:
                os.rename(self.name, self.destination)
            self.delete = False
        except:
            # Delete the file by explicitly calling __del__
            self.__del__()
            self.delete = False  # So the it doesn't try to unlink again
            raise


class TerminatableThread(threading.Thread):
    '''A thread class that supports being terminated from another thread. Updated code from:
    https://github.com/intuited/terminable_thread
    '''
    def terminate(self):
        tid = self.ident
        res = ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_ulong(tid), ctypes.py_object(SystemExit))
        # Check it
        if res == 0:
            raise ValueError("invalid thread id")
        elif res != 1:
            # "if it returns a number greater than one, you're in trouble,
            # and you should call it again with exc=NULL to revert the effect"
            ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tid), None)
            raise SystemError("PyThreadState_SetAsyncExc failed")


def get_open_files():
    open_files = set()
    for p in psutil.process_iter():
        try:
            try:
                func = p.get_open_files
            except AttributeError:
                func = p.open_files

            for f in func():
                open_files.add(f.path)
        except psutil.Error:
            # Usually AccessDenied, but occasionally NoSuchProcess
            pass
    return open_files
