# Copyright (c) 2024 DomainTools LLC
# This work is unpublished, and unauthorized reproduction is prohibited.
# DomainTools, LLC owns all rights to this work and intends that
# it be held confidential, preserved as a trade secret, and protected as
# an unpublished copyrighted work.

import unittest
import tempfile
import os
import shutil
import threading
from time import sleep

from dnstable_manager.cleaner import Cleaner
from dnstable_manager.fileset import Fileset, File, PART_FILE_FORMAT
from dnstable_manager.download import DownloadManager


def gen_part_files(files, td):
    part_file_names = []
    for f in files:
        prefix = PART_FILE_FORMAT.format(f.name)
        tf = tempfile.NamedTemporaryFile(prefix=prefix, dir=td, delete=False)
        part_file_names.append(tf.name)
        tf.close()
    return part_file_names


def mock_get_remote_fileset():
    return set()


class TestCleaner(unittest.TestCase):
    def setUp(self):
        self.td = tempfile.mkdtemp(prefix="test-dnstable-manager-cleaner.")
        self.fileset = Fileset(None, self.td, base="dnssec", extension="mtbl")
        self.fileset.get_remote_fileset = mock_get_remote_fileset
        self.download_manager = DownloadManager()
        self.cleaner = Cleaner(self.fileset, self.download_manager)

    def tearDown(self):
        try:
            shutil.rmtree(self.td)
        except Exception:
            pass

    def test_avoid_single_file(self):
        fname = File("dnssec.20231203.1310.X.mtbl")
        tf = gen_part_files([fname], self.td)[0]
        self.assertTrue(os.path.exists(os.path.join(self.td, tf)))

        self.cleaner._clean()
        self.assertTrue(os.path.exists(os.path.join(self.td, tf)))

    def test_avoid_open(self):
        deca_on_disk = os.path.join(self.td, "dnssec.20231203.1300.X.mtbl")
        with open(deca_on_disk, "w"):
            pass

        fname = "dnssec.20231203.1301.m.mtbl"
        prefix = PART_FILE_FORMAT.format(fname)
        # Create it and keep it open
        tf = tempfile.NamedTemporaryFile(prefix=prefix, dir=self.td, delete=False)

        self.cleaner._clean()
        tf.close()
        self.assertTrue(os.path.exists(os.path.join(self.td, tf.name)))
        self.cleaner._clean()
        self.assertFalse(os.path.exists(os.path.join(self.td, tf.name)))

    def test_avoid_in_manager(self):
        active = {
            File("dnssec.20240221.1920.m.mtbl"): None,
            File("dnssec.20240221.1921.m.mtbl"): None,
            File("dnssec.20240221.1922.m.mtbl"): None,
        }
        self.download_manager._active_downloads = active
        failed = {
            File("dnssec.20240221.1923.m.mtbl"): None
        }
        self.download_manager._failed_downloads = failed
        pending = {File("dnssec.20240221.1923.m.mtbl")}
        self.download_manager._pending_downloads = pending

        expected = []
        expected = gen_part_files(list(active.keys()) + list(failed.keys()) + list(pending), self.td)

        self.cleaner._clean()
        for ef in expected:
            self.assertTrue(os.path.exists(os.path.join(self.td, ef)))

    def test_avoid_not_redundant(self):
        deca_on_disk = os.path.join(self.td, "dnssec.20240221.1910.X.mtbl")
        with open(deca_on_disk, "w"):
            pass
        keep = [
            File("dnssec.20240221.1920.m.mtbl"),
            File("dnssec.20240221.1921.m.mtbl"),
            File("dnssec.20240221.1922.m.mtbl"),
        ]
        redundant = [
            File("dnssec.20240221.1919.m.mtbl"),
            File("dnssec.20240221.1915.m.mtbl"),
        ]

        expected_keep = gen_part_files(keep, self.td)
        expected_delete = gen_part_files(redundant, self.td)

        self.cleaner._clean()

        for ek in expected_keep:
            self.assertTrue(os.path.exists(os.path.join(self.td, ek)))

        for ed in expected_delete:
            self.assertFalse(os.path.exists(os.path.join(self.td, ed)))

    def test_avoid_redundant_full_fileset(self):
        deca_on_disk = os.path.join(self.td, "dnssec.20240221.1910.X.mtbl")
        with open(deca_on_disk, "w"):
            pass
        keep = [
            File("dnssec.20240221.1920.m.mtbl"),
            File("dnssec.20240221.1921.m.mtbl"),
            File("dnssec.20240221.1922.m.mtbl"),
        ]
        redundant = [
            File("dnssec.20240221.1919.m.mtbl"),
            File("dnssec.20240221.1915.m.mtbl"),
        ]

        expected_keep = gen_part_files(keep, self.td) + gen_part_files(redundant, self.td)

        self.cleaner.minimal = False
        self.cleaner._clean()

        for ek in expected_keep:
            self.assertTrue(os.path.exists(os.path.join(self.td, ek)), ek)

    def test_start(self):
        deca_on_disk = os.path.join(self.td, "dnssec.20240221.1910.X.mtbl")
        with open(deca_on_disk, "w"):
            pass

        fname = File("dnssec.20240221.1914.m.mtbl")
        tf = gen_part_files([fname], self.td)[0]
        self.assertTrue(os.path.exists(os.path.join(self.td, tf)))

        self.cleaner.interval = 1
        start_thread = threading.Thread(target=self.cleaner.start, daemon=True)
        start_thread.start()
        sleep(0.2)
        self.assertTrue(os.path.exists(os.path.join(self.td, tf)))
        sleep(1.0)

        self.assertFalse(os.path.exists(os.path.join(self.td, tf)))

    def test_behind(self):
        # Generate part files from a couple hours ago
        outdated = [
            File("dnssec.20240229.1100.H.mtbl"),
            File("dnssec.20240229.1200.X.mtbl"),
            File("dnssec.20240229.1259.m.mtbl"),
        ]
        expected_delete = gen_part_files(outdated, self.td)

        # Simulate a remote that has moved on
        def get_remote_with_data():
            return {
                File("dnssec.20240229.1300.H.mtbl"),
                File("dnssec.20240229.1400.X.mtbl"),
            }
        self.cleaner.fileset.get_remote_fileset = get_remote_with_data

        # Generate part files within or ahead of the remote time period
        keep = [
            File("dnssec.20240229.1300.H.mtbl"),  # An active download
            File("dnssec.20240229.1300.X.mtbl"),  # Redundant by active, but not behind
            File("dnssec.20240229.1401.m.mtbl"),  # Redundant by remote, but not behind
            File("dnssec.20240229.1800.H.mtbl"),  # Way in the future
        ]
        expected_keep = gen_part_files(keep, self.td)

        self.cleaner._clean()
        for ek in expected_keep:
            self.assertTrue(os.path.exists(os.path.join(self.td, ek)), f"{ek}")

        for ed in expected_delete:
            self.assertFalse(os.path.exists(os.path.join(self.td, ed)), f"{ed}")
